import numpy as np
from multiprocessing import Process
import tensorflow as tf
import logging

CHECKPOINT_INTERVAL = 1000000


class ActorLearner(Process):

    def __init__(self, network_creator, environment_creator, args):

        super(ActorLearner, self).__init__()

        self.global_step = 0

        self.max_local_steps = args.max_local_steps
        self.num_actions = args.num_actions
        self.initial_lr = args.initial_lr
        self.lr_pg = args.lr_pg
        self.lr_ce = args.lr_ce
        self.lr_annealing_steps = args.lr_annealing_steps
        self.supervised_batch_size = args.supervised_batch_size
        self.emulator_counts = args.emulator_counts
        self.device = args.device
        self.debugging_folder = args.debugging_folder
        self.network_checkpoint_folder = self.debugging_folder + '/checkpoints/'
        self.optimizer_checkpoint_folder = self.debugging_folder + '/optimizer_checkpoints/'
        self.last_saving_step = 0

        self.emulators = np.asarray([environment_creator.create_environment(i)
                                     for i in range(self.emulator_counts)])
        self.max_global_steps = args.max_global_steps
        self.gamma = args.gamma
        self.game = args.game
        self.network = network_creator()

        self.clip_norm_type = args.clip_norm_type
        self.clip_norm = args.clip_norm
        self.xentropy = args.xentropy

        with tf.device(self.device):
            self.learning_rate = tf.placeholder(tf.float32, shape=[], name='learning_rate')
            self.learning_rate_ce = tf.placeholder(tf.float32, shape=[], name='learning_rate_cross_entropy')
            optimizer_variable_names = 'OptimizerVariables'
            self.optimizer = tf.train.RMSPropOptimizer(self.learning_rate, decay=args.alpha, epsilon=args.e,
                                                       name=optimizer_variable_names)

            self.ce_optimizer = tf.train.RMSPropOptimizer(self.learning_rate_ce, decay=args.alpha, epsilon=args.e,
                                                          name=optimizer_variable_names)
            # Optimizer
            grads_and_vars = self.optimizer.compute_gradients(self.network.loss)

            grads_and_vars = [(g, v) for g, v in grads_and_vars if g is not None]

            self.flat_gradients = tf.concat([tf.reshape(g, [-1]) for g, v in grads_and_vars], axis=0)
            clipped_grads_and_norm = self.clip_gradient([g for g, v in grads_and_vars])
            clipped_grads = clipped_grads_and_norm[0]
            self.grad_norm = clipped_grads_and_norm[1]
            self.clipped_flat_gradients = tf.concat([tf.reshape(g, [-1]) for g in clipped_grads], axis=0)
            self.clipped_grads_and_vars = list(zip(clipped_grads, [v for g, v in grads_and_vars]))

            self.train_step_op = self.optimizer.apply_gradients(self.clipped_grads_and_vars)
            self.optimizer_variables = [var for var in tf.global_variables() if optimizer_variable_names in var.name]

            self.ce_train_step_op_mu = self.ce_optimizer.minimize(self.network.cross_entropy_loss_mu)

            if self.xentropy:
                self.ce_train_step_op = self.ce_optimizer.minimize(self.network.cross_entropy_loss)
            else:
                self.ce_train_step_op = tf.constant(1.0)

        config = tf.ConfigProto()
        if 'gpu' in self.device:
            logging.debug('Dynamic gpu mem allocation')
            config.gpu_options.allow_growth = True

        self.session = tf.Session(config=config)

        self.network_saver = tf.train.Saver(max_to_keep=5)
        self.optimizer_saver = tf.train.Saver(self.optimizer_variables, max_to_keep=5000, name='OptimizerSaver')

    def save_vars(self, force=False):
        if force or self.global_step - self.last_saving_step >= CHECKPOINT_INTERVAL:
            self.last_saving_step = self.global_step
            self.network_saver.save(self.session, self.network_checkpoint_folder, global_step=self.last_saving_step)
            self.optimizer_saver.save(self.session, self.optimizer_checkpoint_folder, global_step=self.last_saving_step)

    def rescale_reward(self, reward):
        """ Clip immediate reward """
        if reward > 1.0:
            reward = 1.0
        elif reward < -1.0:
            reward = -1.0
        return reward

    def init_network(self):
        import os
        if not os.path.exists(self.network_checkpoint_folder):
            os.makedirs(self.network_checkpoint_folder)
        if not os.path.exists(self.optimizer_checkpoint_folder):
            os.makedirs(self.optimizer_checkpoint_folder)

        last_saving_step = self.network.init(self.network_checkpoint_folder, self.network_saver, self.session)

        path = tf.train.latest_checkpoint(self.optimizer_checkpoint_folder)
        if path is not None:
            logging.info('Restoring optimizer variables from previous run')
            self.optimizer_saver.restore(self.session, path)

        return last_saving_step

    def get_lr(self):
        if self.global_step <= self.lr_annealing_steps:
            return self.initial_lr - (self.global_step * self.initial_lr / self.lr_annealing_steps)
        else:
            return 0.0

    def cleanup(self):
        self.save_vars(True)
        self.session.close()

    def clip_gradient(self, grads):
        # This is not really an operation, but a list of gradient Tensors.
        # When calling run() on it, the value of those Tensors
        # (i.e., of the gradients) will be calculated
        if self.clip_norm_type == 'ignore':
            # Unclipped gradients
            gradients = grads
            global_norm = tf.global_norm(gradients)
        elif self.clip_norm_type == 'global':
            # Clip network grads by network norm
            grads_and_norm = tf.clip_by_global_norm(grads, self.clip_norm)
            gradients = grads_and_norm[0]
            # gradients = [v for g, v in gradients if g is not None]
            global_norm = grads_and_norm[1]
        elif self.clip_norm_type == 'local':
            # Clip layer grads by layer norm
            gradients = [tf.clip_by_norm(g, self.clip_norm) for g in grads]
            global_norm = tf.global_norm(gradients)
        else:
            raise Exception('Uh oh')
        return gradients, global_norm

