import tensorflow as tf
import logging
import numpy as np


def flatten(_input):
    shape = _input.get_shape().as_list()
    dim = shape[1]*shape[2]*shape[3]
    return tf.reshape(_input, [-1,dim], name='_flattened')


def conv2d(name, _input, filters, size, channels, stride, padding = 'VALID', init = "torch"):
    w = conv_weight_variable([size,size,channels,filters],
                             name + '_weights', init = init)
    b = conv_bias_variable([filters], size, size, channels,
                           name + '_biases', init = init)
    conv = tf.nn.conv2d(_input, w, strides=[1, stride, stride, 1],
                        padding=padding, name=name + '_convs')
    out = tf.nn.relu(tf.add(conv, b),
                     name='' + name + '_activations')
    return w, b, out


def conv_weight_variable(shape, name, init = "torch"):
    if init == "glorot_uniform":
        receptive_field_size = np.prod(shape[:2])
        fan_in = shape[-2] * receptive_field_size
        fan_out = shape[-1] * receptive_field_size
        d = np.sqrt(6. / (fan_in + fan_out))
    else:
        w = shape[0]
        h = shape[1]
        input_channels = shape[2]
        d = 1.0 / np.sqrt(input_channels * w * h)

    initial = tf.random_uniform(shape, minval=-d, maxval=d)
    return tf.Variable(initial, name=name, dtype='float32')


def conv_bias_variable(shape, w, h, input_channels, name, init= "torch"):
    if init == "glorot_uniform":
        initial = tf.zeros(shape)
    else:
        d = 1.0 / np.sqrt(input_channels * w * h)
        initial = tf.random_uniform(shape, minval=-d, maxval=d)
    return tf.Variable(initial, name=name, dtype='float32')


def fc(name, _input, output_dim, activation = "relu", init = "torch"):
    input_dim = _input.get_shape().as_list()[1]
    w = fc_weight_variable([input_dim, output_dim],
                           name + '_weights', init = init)
    b = fc_bias_variable([output_dim], input_dim,
                         '' + name + '_biases', init = init)
    out = tf.add(tf.matmul(_input, w), b, name= name + '_out')

    if activation == "relu":
        out = tf.nn.relu(out, name='' + name + '_relu')

    return w, b, out


def fc_weight_variable(shape, name, init="torch"):
    if init == "glorot_uniform":
        fan_in = shape[0]
        fan_out = shape[1]
        d = np.sqrt(6. / (fan_in + fan_out))
    else:
        input_channels = shape[0]
        d = 1.0 / np.sqrt(input_channels)
    initial = tf.random_uniform(shape, minval=-d, maxval=d)
    return tf.Variable(initial, name=name, dtype='float32')


def fc_bias_variable(shape, input_channels, name, init= "torch"):
    if init=="glorot_uniform":
        initial = tf.zeros(shape, dtype='float32')
    else:
        d = 1.0 / np.sqrt(input_channels)
        initial = tf.random_uniform(shape, minval=-d, maxval=d)
    return tf.Variable(initial, name=name, dtype='float32')


def linear(name, _input, output_dim):
    input_dim = _input.get_shape().as_list()[1]
    w = fc_weight_variable([input_dim, output_dim], name + '_weights')
    b = fc_bias_variable([output_dim], input_dim, name + '_biases')
    out = tf.add(tf.matmul(_input, w), b, name= name + '_out')
    return w, b, out


def softmax(name, _input, output_dim):
    input_dim = _input.get_shape().as_list()[1]
    w = fc_weight_variable([input_dim, output_dim], name + '_weights')
    b = fc_bias_variable([output_dim], input_dim, name + '_biases')
    out = tf.nn.softmax(tf.add(tf.matmul(_input, w), b), name= name + '_policy')
    return w, b, out


def log_softmax( name, _input, output_dim):
    input_dim = _input.get_shape().as_list()[1]
    w = fc_weight_variable([input_dim, output_dim], name + '_weights')
    b = fc_bias_variable([output_dim], input_dim, name + '_biases')
    out = tf.nn.log_softmax(tf.add(tf.matmul(_input, w), b), name= name + '_policy')
    return w, b, out


class Network(object):

    def __init__(self, conf):
        """ Initialize hyper-parameters, set up optimizer and network
        layers common across Q and Policy/V nets. """

        self.name = conf['name']
        self.num_actions = conf['num_actions']
        self.emulator_counts = conf['emulator_counts']
        self.device = conf['device']

        with tf.device(self.device):
            with tf.name_scope(self.name):
                self.loss_scaling = 5.0
                self.input_ph = tf.placeholder(tf.uint8, [None, 84, 84, 4], name='input')
                self.selected_action_ph = tf.placeholder("float32", [None, self.num_actions], name="selected_action")
                self.input = tf.scalar_mul(1.0/255.0, tf.cast(self.input_ph, tf.float32))

                # This class should never be used, must be subclassed

                # The output layer
                self.output = None

    def init(self, checkpoint_folder, saver, session):
        last_saving_step = 0

        with tf.device('/cpu:0'):
            # Initialize network parameters
            path = tf.train.latest_checkpoint(checkpoint_folder)
            if path is None:
                logging.info('Initializing all variables')
                session.run(tf.global_variables_initializer())
            else:
                logging.info('Restoring network variables from previous run')
                saver.restore(session, path)
                last_saving_step = int(path[path.rindex('-')+1:])
        return last_saving_step


class NIPSNetwork(Network):

    def __init__(self, conf):
        super(NIPSNetwork, self).__init__(conf)

        with tf.device(self.device):
            with tf.name_scope(self.name):
                _, _, conv1 = conv2d('conv1', self.input, 16, 8, 4, 4)

                _, _, conv2 = conv2d('conv2', conv1, 32, 4, 16, 2)

                _, _, fc3 = fc('fc3', flatten(conv2), 256, activation="relu")

                self.output = fc3


class NatureNetwork(Network):

    def __init__(self, conf):
        super(NatureNetwork, self).__init__(conf)

        with tf.device(self.device):
            with tf.name_scope(self.name):
                _, _, conv1 = conv2d('conv1', self.input, 32, 8, 4, 4)

                _, _, conv2 = conv2d('conv2', conv1, 64, 4, 32, 2)

                _, _, conv3 = conv2d('conv3', conv2, 64, 3, 64, 1)

                _, _, fc4 = fc('fc4', flatten(conv3), 512, activation="relu")

                self.output = fc4


class LSTMNetwork(Network):

    def __init__(self, conf):
        super(LSTMNetwork, self).__init__(conf)

        self.lstm_size = conf['lstm_size']
        self.lstm_state = (tf.zeros((self.emulator_counts, self.lstm_size), dtype=tf.float32),
                           (tf.zeros((self.emulator_counts, self.lstm_size), dtype=tf.float32)))

        with tf.device(self.device):
            with tf.name_scope(self.name):
                # 0.0 if the episode was over on the previous timestep, else 1.0
                self.prev_episode_over_mask_ph = tf.placeholder(tf.float32, [None, self.emulator_counts], name='episode_over_mask')
                self.lstm_state_c = tf.Variable(self.lstm_state[0], trainable=False, dtype=tf.float32)
                self.lstm_state_h = tf.Variable(self.lstm_state[1], trainable=False, dtype=tf.float32)

                self.reset_state = tf.group(tf.assign(self.lstm_state_c, self.lstm_state[0]),
                                            tf.assign(self.lstm_state_h, self.lstm_state[1]))

                stored_lstm_state_c = tf.Variable(self.lstm_state[0], trainable=False, dtype=tf.float32)
                stored_lstm_state_h = tf.Variable(self.lstm_state[1], trainable=False, dtype=tf.float32)

                stored_lstm_state = tf.contrib.rnn.LSTMStateTuple(self.lstm_state_c, self.lstm_state_h)

                input_dim = self.output.get_shape().as_list()[1]

                reshaped_input = tf.reshape(self.output, [-1, self.emulator_counts, input_dim])

                max_time = tf.cast(tf.gather(tf.cast(tf.shape(reshaped_input), dtype=tf.float32), 0), dtype=tf.int32)
                self.max_time = max_time

                inputs_ta = tf.TensorArray(dtype=tf.float32, size=1, dynamic_size=True, name='tensor_array_inputs', clear_after_read=False)
                inputs_ta = inputs_ta.unstack(reshaped_input, 'unstack_inputs')
                episode_over_ta = tf.TensorArray(dtype=tf.float32, size=self.emulator_counts, dynamic_size=True, name='tensor_array_episode_over')
                episode_over_ta = episode_over_ta.unstack(self.prev_episode_over_mask_ph, 'unstack_episode_over_mask')

                cell = tf.contrib.rnn.BasicLSTMCell(self.lstm_size)

                def loop_fn(time, cell_output, cell_state, loop_state):
                    nonlocal max_time
                    emit_output = cell_output  # == None for time == 0

                    if cell_output is None:  # time == 0
                        use_cell_state = stored_lstm_state
                    else:
                        use_cell_state = cell_state

                    episode_over = tf.expand_dims(episode_over_ta.read(time), axis=1)
                    # If episode ended, next cell state should be zero.
                    next_cell_state = tf.contrib.rnn.LSTMStateTuple(tf.multiply(use_cell_state[0], episode_over),
                                                                    tf.multiply(use_cell_state[1], episode_over))
                    elements_finished = tf.greater_equal(time, max_time, name='elements_finished')
                    # If time = max_time then the loop will finish and next_input will not be used,
                    # this minimum effectively replaces a conditional statement by reusing the last input.
                    next_input = inputs_ta.read(tf.minimum(time, max_time-1))
                    next_loop_state = None
                    return (elements_finished, next_input, next_cell_state,
                            emit_output, next_loop_state)

                outputs_ta, final_state, _ = tf.nn.raw_rnn(cell, loop_fn)

                update_lstm_state = tf.group(tf.assign(self.lstm_state_c, final_state[0]),
                                             tf.assign(self.lstm_state_h, final_state[1]))
                with tf.control_dependencies([update_lstm_state]):
                    self.output_stack = outputs_ta.stack()

                self.output = tf.reshape(self.output_stack, [-1, self.lstm_size])

                self.store_lstm_state = tf.group(tf.assign(stored_lstm_state_c, self.lstm_state_c),
                                                 tf.assign(stored_lstm_state_h, self.lstm_state_h))

                self.rollback_lstm_state = tf.group(tf.assign(self.lstm_state_c, stored_lstm_state_c),
                                                    tf.assign(self.lstm_state_h, stored_lstm_state_h))

                # Normal flow
                # 1. store_lstm_state
                # 2. output + update_lstm_state
                # 3. repeat 2 max_local_steps timesteps
                # 4. output + update_lstm_state
                # 5. rollback_lstm_state
                # 6. output + update weights + update_lstm_state


class LSTMNatureVNetwork(LSTMNetwork, NatureNetwork):
    pass


class LSTMNipsVNetwork(LSTMNetwork, NIPSNetwork):
    pass