import time
from multiprocessing.sharedctypes import RawArray
from ctypes import c_uint, c_float
from actor_learner import *
import logging
import os

from circular_buffer_index import CircularBufferIndex

from emulator_runner import EmulatorRunner
from replay_memory import ReplayMemory, get_buffers_for_arrays
from runners import Runners
import numpy as np

np.set_printoptions(threshold=np.inf)
import custom_logging


class PAACLearner(ActorLearner):
    def __init__(self, network_creator, environment_creator, args):
        super(PAACLearner, self).__init__(network_creator, environment_creator, args)
        self.workers = args.emulator_workers
        actions_header = ','.join(environment_creator.action_strings)
        logger_config = [{'name': 'environment', 'file_name': 'env_log.txt',
                          'header': 'Relative time|Absolute time|Global Time Step|Environment nr|Steps to finish|Reward|[{}]'.format(
                              actions_header)},
                         {'name': 'learning', 'file_name': 'learn_log.txt',
                          'header': 'Relative time|Absolute time|Global Time Step|[R]|[Value]|[RawG]|[CorrG]|[PG norm, Xent norm]'},
                         {'name': 'supervised', 'file_name': 'supervised_log.txt',
                          'header': 'Relative time|Absolute time|Global Time Step|[Gradients]|[Logits]|[LogitsDiff]|[LogitsGrad]'},
                         ]
        self.stats_logger = custom_logging.StatsLogger(logger_config, subfolder=args.debugging_folder)
        self.memory_size = args.memory_size
        self.memory_updates = args.memory_updates

        # Dummy to allow the use of optimizers in partial_run
        with tf.device(self.device):
            with tf.control_dependencies([self.train_step_op]):
                self.train_step_op = tf.constant(0, name='train_step_op')

    @staticmethod
    def choose_next_actions(network, num_actions, states, session):
        network_output_v, network_output_pi = session.run(
            [network.output_layer_v,
             network.output_layer_pi],
            feed_dict={network.input_ph: states})

        action_indices = PAACLearner.__sample_policy_action(network_output_pi)

        new_actions = np.eye(num_actions)[action_indices]

        return new_actions, network_output_v, network_output_pi

    def __choose_next_actions(self, states):
        return PAACLearner.choose_next_actions(self.network, self.num_actions, states, self.session)

    @staticmethod
    def __sample_policy_action(probs):
        """
        Sample an action from an action probability distribution output by
        the policy network.
        """
        # Subtract a tiny value from probabilities in order to avoid
        # "ValueError: sum(pvals[:-1]) > 1.0" in numpy.multinomial
        probs = probs - np.finfo(np.float32).epsneg

        action_indexes = [int(np.nonzero(np.random.multinomial(1, p))[0]) for p in probs]
        return action_indexes

    def _get_shared(self, array, dtype=c_float):
        """
        Returns a RawArray backed numpy array that can be shared between processes.
        :param array: the array to be shared
        :param dtype: the RawArray dtype to use
        :return: the RawArray backed numpy array
        """

        shape = array.shape
        shared = RawArray(dtype, array.reshape(-1))
        return np.frombuffer(shared, dtype).reshape(shape)

    def train(self):
        """
        Main actor learner loop for parallel advantage actor critic learning.
        """
        updates = 0
        self.global_step = self.init_network()

        logging.debug("Starting training at Step {}".format(self.global_step))
        counter = 0

        total_rewards = []

        # state, reward, episode_over, action
        variables = [(np.asarray([emulator.get_initial_state() for emulator in self.emulators], dtype=np.uint8)),
                     (np.zeros(self.emulator_counts, dtype=np.float32)),
                     (np.asarray([False] * self.emulator_counts, dtype=np.float32)),
                     (np.zeros((self.emulator_counts, self.num_actions), dtype=np.float32))]

        self.runners = Runners(EmulatorRunner, self.emulators, self.workers, variables)
        self.runners.start()
        shared_states, shared_rewards, shared_episode_over, shared_actions = self.runners.get_shared_variables()

        action_probs_sum = np.zeros((self.num_actions))

        emulator_steps = [0] * self.emulator_counts

        total_episode_rewards = self.emulator_counts * [0]

        rewards = np.zeros((self.max_local_steps, self.emulator_counts), dtype=np.float32)
        states = np.zeros([self.max_local_steps] + list(shared_states.shape), dtype=np.uint8)
        actions = np.zeros((self.max_local_steps, self.emulator_counts, self.num_actions), dtype=np.float32)
        action_probabilities = np.zeros((self.max_local_steps, self.emulator_counts, self.num_actions),
                                        dtype=np.float32)
        episodes_over_masks = np.zeros((self.max_local_steps, self.emulator_counts), dtype=np.float32)

        if self.xentropy:
            self.replay_memory = ReplayMemory(get_buffers_for_arrays([states[0][0],
                                                                      actions[0][0],
                                                                      action_probabilities[0][0],
                                                                      rewards[0][0],
                                                                      episodes_over_masks[0][0],
                                                                      action_probabilities[0][0],
                                                                      ],
                                                                     self.memory_size))
        else:
            self.replay_memory = ReplayMemory(get_buffers_for_arrays([states[0],
                                                                      actions[0],
                                                                      action_probabilities[0],
                                                                      rewards[0],
                                                                      episodes_over_masks[0],
                                                                      ],
                                                                     self.memory_size))


        circular_buffer_index = CircularBufferIndex(self.max_local_steps)
        true_timestep = 0
        prev_global_step = self.global_step

        self.learning_summaries = [self.flat_gradients, self.clipped_flat_gradients, self.grad_norm, self.network.xentropy_grads_norm]
        # self.learning_summaries = [tf.constant(1.0), tf.constant(1.0), tf.constant(1.0), tf.constant(1.0)]
        self.learning_summaries_mu = [self.network.xentropy_grads_mu, self.network.flat_logits, self.network.logits_diff, self.network.xentropy_logit_grads_mu]
        # self.learning_summaries_mu = [tf.constant(1.0), tf.constant(1.0), tf.constant(1.0), tf.constant(1.0)]

        start_time = time.time()

        while self.global_step < self.max_global_steps:

            loop_start_time = time.time()

            for _ in range(max(self.max_local_steps, self.max_local_steps - circular_buffer_index.current_size)):
                t = circular_buffer_index.current_index
                next_actions, readouts_v_t, readouts_pi_t = self.__choose_next_actions(shared_states)
                action_probs_sum += np.mean(readouts_pi_t, axis=0)
                for z in range(next_actions.shape[0]):
                    shared_actions[z] = next_actions[z]

                actions[t] = next_actions
                action_probabilities[t] = readouts_pi_t
                states[t] = shared_states

                # Start updating all environments with next_actions
                self.runners.update_environments()
                self.runners.wait_updated()
                # Done updating all environments, have new states, rewards and is_over

                episodes_over_masks[t] = 1.0 - shared_episode_over.astype(np.float32)

                some_episode_over = False
                avg_action_probs = action_probs_sum / np.sum(action_probs_sum)

                for e, (actual_reward, episode_over) in enumerate(zip(shared_rewards, shared_episode_over)):
                    total_episode_rewards[e] += actual_reward
                    actual_reward = self.rescale_reward(actual_reward)
                    rewards[t, e] = actual_reward

                    emulator_steps[e] += 1
                    self.global_step += 1
                    if episode_over:
                        some_episode_over = True
                        total_rewards.append(total_episode_rewards[e])
                        self.stats_logger.log('environment', self.global_step, e, emulator_steps[e],
                                              total_episode_rewards[e], avg_action_probs)
                        total_episode_rewards[e] = 0
                        emulator_steps[e] = 0
                if some_episode_over:
                    action_probs_sum = np.zeros(self.num_actions)
                true_timestep += 1
                circular_buffer_index.increase()

            flat_critic_targets, next_state_values, summaries = self.update(states, actions, rewards, shared_states,
                                                                            action_probabilities,
                                                                            episodes_over_masks,
                                                                            list(reversed(circular_buffer_index.get_all_indices())))
            updates += 1
            self.stats_logger.log('learning', self.global_step,
                                  self.stats_logger.get_stats_for_array(flat_critic_targets),
                                  self.stats_logger.get_stats_for_array(next_state_values),
                                  self.stats_logger.get_stats_for_array(summaries[0]),
                                  self.stats_logger.get_stats_for_array(summaries[1]),
                                  [summaries[2], summaries[3]],
                                  )

            if self.memory_updates > 0 and counter > 10:
                if self.xentropy:
                    self.xent_off_policy_update()
                else:
                    self.pg_off_policy_update()
                updates += self.memory_updates

            counter += 1
            if counter % (2048 / self.emulator_counts) == 0:
                curr_time = time.time()
                global_steps = self.global_step
                last_ten = 0.0 if len(total_rewards) < 1 else np.mean(total_rewards[-10:])
                updates_per_sec = updates / (curr_time - start_time)
                true_timesteps_per_sec = true_timestep / (curr_time - start_time)
                average_environment_timesteps_per_sec = global_steps / (curr_time - start_time)
                current_environment_timesteps_per_sec = (global_steps - prev_global_step) / (
                    curr_time - loop_start_time)
                logging.info(
                    "Step {}, {:0.2f} updates/s, {:0.2f} timesteps/s, {:0.2f} env steps/s (current {:0.2f} env steps/s), last 10 rewards avg {:0.2f}"
                        .format(global_steps,
                                updates_per_sec,
                                true_timesteps_per_sec,
                                average_environment_timesteps_per_sec,
                                current_environment_timesteps_per_sec,
                                last_ten))
            prev_global_step = self.global_step
            self.save_vars()
        self.cleanup()

    def update(self, states, actions, rewards, next_states, action_probabilities, episodes_over_masks, indices, off_policy=False):
        timesteps = states.shape[0]
        lr = self.get_lr()
        train_step_op = self.train_step_op
        train_step_op_ce = self.ce_train_step_op


        critic_targets = np.empty((timesteps, self.emulator_counts))
        next_state_values = self.session.run(
            self.network.output_layer_v,
            feed_dict={self.network.input_ph: next_states})
        flat_states = states.reshape([timesteps * self.emulator_counts] + list(next_states.shape)[1:])
        flat_actions = actions.reshape(timesteps * self.emulator_counts, self.num_actions)
        flat_action_probs = action_probabilities.reshape(timesteps * self.emulator_counts, self.num_actions)

        if off_policy:
            feeds = [self.network.input_ph, self.network.critic_target_ph, self.network.selected_action_ph,
                     self.network.action_probabilities_ph, self.learning_rate, self.network.learning_rate_pg]
            setup = self.session.partial_run_setup([self.network.selected_action_is,
                                                    train_step_op,
                                                    self.learning_summaries], feeds)
            batch_selected_is = self.session.partial_run(setup, self.network.selected_action_is,
                                                         feed_dict={self.network.input_ph: flat_states,
                                                                    self.network.selected_action_ph: flat_actions,
                                                                    self.network.action_probabilities_ph: flat_action_probs})
            # Following ACER paper c' = 1
            clipped_batch_selected_is = np.clip(batch_selected_is, 0.0, 1.0)
            Rs = np.zeros_like(next_state_values)
            discounted_v = next_state_values
            for t in indices:
                discounted_v = self.gamma * discounted_v * episodes_over_masks[t]
                Rs = clipped_batch_selected_is[t] * (rewards[t] + self.gamma * Rs * episodes_over_masks[t])
                n_step_approx = Rs + discounted_v
                critic_targets[t] = np.copy(n_step_approx)

            flat_critic_targets = critic_targets.reshape(-1)
            feed_dict = {self.network.critic_target_ph: flat_critic_targets,
                         self.learning_rate: lr,
                         self.network.learning_rate_pg: self.lr_pg}
            _, *summaries = self.session.partial_run(setup, [train_step_op] + self.learning_summaries,
                                                     feed_dict=feed_dict)
        else:
            Rs = np.zeros_like(next_state_values)
            discounted_v = next_state_values
            for t in indices:
                discounted_v = self.gamma * discounted_v * episodes_over_masks[t]
                Rs = rewards[t] + self.gamma * Rs * episodes_over_masks[t]
                n_step_approx = Rs + discounted_v
                critic_targets[t] = np.copy(n_step_approx)

            flat_critic_targets = critic_targets.reshape(-1)
            flat_rewards = rewards.reshape(-1)
            flat_episodes_over_masks = episodes_over_masks.reshape(-1)

            feed_dict = {self.network.input_ph: flat_states,
                         self.network.selected_action_ph: flat_actions,
                         self.network.critic_target_ph: flat_critic_targets,
                         self.network.action_probabilities_ph: flat_action_probs,
                         self.learning_rate: lr,
                         self.learning_rate_ce: self.lr_ce,
                         self.network.learning_rate_pg: self.lr_pg}

            _, _, target_policies,  *summaries = self.session.run(
                [train_step_op, train_step_op_ce, self.network.target_policy,] + self.learning_summaries, feed_dict=feed_dict)

            flat_target_policies = target_policies.reshape(self.max_local_steps * self.emulator_counts,
                                                           self.num_actions)

            if self.memory_updates > 0:
                if self.xentropy:
                    self.replay_memory.add(
                        (flat_states, flat_actions, flat_action_probs, flat_rewards, flat_episodes_over_masks,
                         flat_target_policies))
                else:
                    self.replay_memory.add((states, actions, action_probabilities, rewards, episodes_over_masks))

        return flat_critic_targets, next_state_values, summaries

    def xent_off_policy_update(self):
        __all_summaries = []
        for i in range(self.memory_updates):
            _states, _actions, _action_probs, _rewards, _, _target_policies = self.replay_memory.sample(
                self.supervised_batch_size)
            feed_dict = {self.network.input_ph: _states,
                         self.network.target_policy_ph: _target_policies,
                         self.learning_rate_ce: self.lr_ce/10.0,
                         self.network.learning_rate_pg: self.lr_pg}
            _, *_summaries = self.session.run([self.ce_train_step_op_mu] + self.learning_summaries_mu,
                                              feed_dict=feed_dict)
            concatenate = np.concatenate(_summaries[0])
            #TODO: Remove
            if np.max(concatenate) > 1:
                # self.network.xentropy_grads_mu, self.network.flat_logits, self.network.logits_diff, self.network.xentropy_logit_grads_mu, _actions, _action_probs, _rewardss, _target_policiess
                for j, _s in enumerate(_summaries + [_actions, _action_probs, _rewards, _target_policies]):
                    np.save(os.path.join(self.debugging_folder, 'array_{}_{}.npy'.format(self.global_step, j)), _s)
            __all_summaries.append([concatenate] + _summaries[1:])
        __all_summaries = list(zip(*__all_summaries))
        self.stats_logger.log('supervised', self.global_step,
                              self.stats_logger.get_stats_for_array(__all_summaries[0]),
                              self.stats_logger.get_stats_for_array(__all_summaries[1]),
                              self.stats_logger.get_stats_for_array(__all_summaries[2]),
                              self.stats_logger.get_stats_for_array(__all_summaries[3]),
                              )

    def pg_off_policy_update(self):
        __all_summaries = []
        for i in range(self.memory_updates):
            timesteps = self.supervised_batch_size // self.emulator_counts
            __states, _actions, _action_probs, _rewards, _episodes_over_masks = self.replay_memory.sample_contiguous(
                timesteps + 1)
            _states = __states[:-1]
            _next_states = __states[-1]
            _actions = _actions[:-1]
            _action_probs = _action_probs[:-1]
            _rewards = _rewards[:-1]
            _episodes_over_masks = _episodes_over_masks[:-1]
            flat_critic_targets, next_state_values, _summaries = self.update(_states, _actions, _rewards,
                                                                             _next_states, _action_probs,
                                                                             _episodes_over_masks,
                                                                             list(reversed(range(timesteps))),
                                                                             off_policy=True)

            __all_summaries.append([flat_critic_targets, next_state_values] + _summaries)
        __all_summaries = list(zip(*__all_summaries))
        self.stats_logger.log('supervised', self.global_step,
                              self.stats_logger.get_stats_for_array(__all_summaries[0]),
                              self.stats_logger.get_stats_for_array(__all_summaries[1]),
                              self.stats_logger.get_stats_for_array(__all_summaries[-2]),
                              self.stats_logger.get_stats_for_array(__all_summaries[-1]),
                              )

    def cleanup(self):
        super(PAACLearner, self).cleanup()
        self.runners.stop()
