from networks import *


class PolicyVNetwork(Network):

    def __init__(self, conf):
        """ Set up remaining layers, objective and loss functions, gradient
        compute and apply ops, network parameter synchronization ops, and
        summary ops. """

        super(PolicyVNetwork, self).__init__(conf)
        self.xentropy = conf['xentropy']

        self.entropy_regularisation_strength = conf['entropy_regularisation_strength']

        # Toggle additional recurrent layer
        with tf.device(conf['device']):
            with tf.name_scope(self.name):
                self.learning_rate_pg = tf.placeholder(tf.float32, shape=[], name='learning_rate_policy_gradient')
                self.action_probabilities_ph = tf.placeholder("float32", [None, self.num_actions], name="action_probabilities")

                self.critic_target_ph = tf.placeholder(
                    "float32", [None], name='target')
                self.target_policy_ph = tf.placeholder("float32", [None, self.num_actions], name="target_policy_ph")

                layer_name = 'actor_output'

                _, _, self.output_layer_pi_logits = linear(layer_name, self.output, self.num_actions)

                self.flat_logits = tf.reshape(self.output_layer_pi_logits, [-1])

                self.logits_diff = tf.reduce_max(self.output_layer_pi_logits, axis=1) - tf.reduce_min(self.output_layer_pi_logits, axis=1)

                # Final actor layer
                self.output_layer_pi = tf.nn.softmax(self.output_layer_pi_logits, name=layer_name)
                # Final critic layer
                _, _, self.output_layer_v = fc('critic_output', self.output, 1, activation="linear")

                importance_sampling = tf.stop_gradient(self.output_layer_pi) / tf.add(self.action_probabilities_ph, tf.constant(1e-10))
                self.selected_action_is = tf.reduce_sum(
                    tf.multiply(importance_sampling, self.selected_action_ph),
                    reduction_indices=1, name='selected_action_is')
                clipped_selected_action_is = tf.clip_by_value(self.selected_action_is, 0.0, 10.0)

                # Avoiding log(0) by adding a very small quantity (1e-20) to output.
                self.log_output_layer_pi = tf.log(tf.add(self.output_layer_pi, tf.constant(1e-20)),
                                                  name=layer_name + '_log_policy')

                # Entropy: sum_a (-p_a ln p_a)
                self.output_layer_entropy = tf.reduce_sum(tf.multiply(
                    tf.constant(-1.0),
                    tf.multiply(self.output_layer_pi, self.log_output_layer_pi)), reduction_indices=1)

                self.output_layer_v = tf.reshape(self.output_layer_v, [-1])

                self.critic_error = tf.subtract(self.critic_target_ph, self.output_layer_v)

                self.actor_advantage = tf.stop_gradient(self.critic_error)

                # Actor objective
                # Multiply the output of the network by a one hot vector, 1 for the
                # executed action. This will make the non-regularised objective
                # term for non-selected actions to be zero.
                log_output_selected_action = tf.reduce_sum(
                    tf.multiply(self.log_output_layer_pi, self.selected_action_ph),
                    reduction_indices=1)
                self.actor_objective_advantage_term = tf.multiply(
                    log_output_selected_action, self.actor_advantage)
                self.actor_objective_entropy_term = tf.multiply(
                    self.entropy_regularisation_strength, self.output_layer_entropy)

                self.is_corrected_actor_objective_advantage_term = tf.multiply(clipped_selected_action_is,
                                                                               self.actor_objective_advantage_term)

                actor_objective = tf.multiply(tf.constant(-1.0), self.is_corrected_actor_objective_advantage_term)
                self.actor_objective_mean = tf.reduce_mean(actor_objective, name='mean_actor_objective')

                actor_entropy = tf.multiply(tf.constant(-1.0), self.actor_objective_entropy_term)
                self.actor_entropy_mean = tf.reduce_mean(actor_entropy, name='mean_actor_entropy')

                critic_loss = tf.scalar_mul(0.25, tf.pow(self.critic_error, 2))
                self.critic_loss_mean = tf.reduce_mean(critic_loss, name='mean_critic_loss')

                self.abs_per_example_loss = tf.abs(actor_objective + critic_loss)

                pg_loss = actor_objective + actor_entropy
                sum_pg_loss = tf.reduce_sum(pg_loss, name='sum_actor_loss')
                policy_gradient = tf.gradients(sum_pg_loss, self.output_layer_pi_logits)[0]
                target_policy = tf.nn.softmax(self.output_layer_pi_logits - self.learning_rate_pg * policy_gradient)
                self.target_policy = tf.stop_gradient(target_policy)

                self.target_policy_cross_entropy = tf.nn.softmax_cross_entropy_with_logits(logits=self.output_layer_pi_logits, labels=self.target_policy)

                self.target_policy_cross_entropy_mu = tf.nn.softmax_cross_entropy_with_logits(logits=self.output_layer_pi_logits, labels=self.target_policy_ph)

                if self.xentropy:
                    self.loss = self.critic_loss_mean
                else:
                    self.loss = self.critic_loss_mean + self.actor_objective_mean + self.actor_entropy_mean

                self.loss = tf.multiply(self.loss, self.loss_scaling, 'loss')
                self.cross_entropy_loss = tf.reduce_mean(self.target_policy_cross_entropy, name='cross_entropy_loss')
                self.cross_entropy_loss_mu = tf.reduce_mean(self.target_policy_cross_entropy_mu, name='cross_entropy_loss_experience_replay')

                self.xentropy_logit_grads_mu = self.output_layer_pi - self.target_policy_ph

                network_vars = [var for var in tf.trainable_variables() if self.name in var.name]
                self.actor_grads = [tf.reshape(g, [-1]) for g in tf.gradients(self.actor_objective_mean, network_vars) if g is not None]
                self.entropy_grads = [tf.reshape(g, [-1]) for g in tf.gradients(self.actor_entropy_mean, network_vars) if g is not None]
                self.critic_grads = [tf.reshape(g, [-1]) for g in tf.gradients(self.critic_loss_mean, network_vars) if g is not None]
                self.xentropy_grads = [tf.reshape(g, [-1]) for g in tf.gradients(self.cross_entropy_loss, network_vars) if g is not None]
                self.xentropy_grads_mu = [tf.reshape(g, [-1]) for g in tf.gradients(self.cross_entropy_loss_mu, network_vars) if g is not None]
                self.diff_grads = tf.concat([ag - xg for ag, xg in zip(self.actor_grads, self.xentropy_grads)], axis=0)

                self.xentropy_grads_norm = tf.global_norm(self.xentropy_grads)

                self.actor_grads = tf.abs(tf.concat(self.actor_grads, axis=0))
                self.entropy_grads = tf.abs(tf.concat(self.entropy_grads, axis=0))
                self.critic_grads = tf.abs(tf.concat(self.critic_grads, axis=0))
                self.xentropy_grads = tf.concat(self.xentropy_grads, axis=0)
                self.flat_xentropy_grads_mu = tf.concat(self.xentropy_grads_mu, axis=0)



class NIPSPolicyVNetwork(PolicyVNetwork, NIPSNetwork):
    pass


class NaturePolicyVNetwork(PolicyVNetwork, NatureNetwork):
    pass
