import numpy as np
import itertools


def get_buffers_for_arrays(arrays, buffer_size):
    """
    Helper method to get buffers for the given arrays, to be used with the constructor.
    :param arrays: the arrays to create buffers for.
    :param buffer_size: the size of the buffers.
    :return: the buffers.
    """
    return [np.empty([buffer_size] + list(array.shape), dtype=array.dtype) for array in arrays]


class ReplayMemory(object):

    def __init__(self, buffers):
        """
        Initializes the replay memory
        :param buffers: list of buffers. For example [np.empty((memory_size, width, height, depth)), np.empty(memory_size)]
        """

        if len({b.shape[0] for b in buffers}) != 1:
            raise Exception('All buffers must contain the same amount of rows.')

        self.buffers = buffers
        self.memory_size = buffers[0].shape[0]
        self.current = 0
        self.count = 0

    def add(self, data):
        """
        Adds data to the replay memory
        :param data: list of data to be added. For example [np.ones((10, width, height, depth)), np.zeros(10)]
        """
        if len(data) != len(self.buffers):
            raise Exception('Must supply data for all buffers at once')
        data_size = data[0].shape[0]
        indices = self.__get_insertion_indices(data_size)
        for d, b in zip(data, self.buffers):
            b[indices] = np.copy(d)
        self.count = min(self.memory_size, self.count + data_size)
        self.current = (self.current + data_size) % self.memory_size

    def sample(self, count):
        """
        Returns count randomly sampled memories
        :param count:
        :return:
        """
        if count > self.count:
            raise Exception('Not enough memories in buffer')
        indices = np.random.randint(0, self.count, count)
        return [x[indices] for x in self.buffers]

    def sample_last_contiguous(self, count, skip=0):

        count_skip = count + skip
        if count_skip > self.count:
            raise Exception('Not enough memories in buffer')
        start = self.current - 1 - count_skip if self.current - 1 - count_skip >= 0 else self.count - count_skip + self.current
        indices = self.__get_contiguous_indices(start, count)
        return [x[indices] for x in self.buffers]

    # TODO: Maybe modify this function so that each environment can be sampled independently (may better decorrelate samples)
    def sample_contiguous(self, count):
        """
        Returns count randomly sampled contiguous memories
        :param count:
        :return:
        """
        if count > self.count:
            raise Exception('Not enough memories in buffer')
        start = np.random.randint(0, self.count - count, 1)[0]
        indices = self.__get_contiguous_indices(start, count)
        return [x[indices] for x in self.buffers]

    def __get_insertion_indices(self, data_size):
        return self.__get_contiguous_indices(self.current, data_size)

    def __get_contiguous_indices(self, start, count):
        if start + count > self.memory_size:
            return list(range(start, self.memory_size)) + list(range(count-(self.memory_size - start)))
        else:
            return list(range(start, start + count))


class PrioritizedReplayMemory(object):

    def __init__(self, buffers):
        """
        Initializes the replay memory
        :param buffers: list of buffers. For example [np.empty((memory_size, width, height, depth)), np.empty(memory_size)]
        """

        if len({b.shape[0] for b in buffers}) != 1:
            raise Exception('All buffers must contain the same first and second dimensions.')

        self.buffers = buffers
        self.memory_size = buffers[0].shape[0]
        self.priorities = np.zeros(self.memory_size, dtype=np.float32)
        self.current = 0
        self.count = 0

    def add(self, data, priorities):
        """
        Adds data to the replay memory
        :param data: list of data to be added. For example [np.ones((10, width, height, depth)), np.zeros(10)]
        """
        if len(data) != len(self.buffers):
            raise Exception('Must supply data for all buffers at once')
        data_size = data[0].shape[0]
        indices = self.__get_insertion_indices(data_size)
        for d, b in zip(data, self.buffers):
            b[indices] = np.copy(d)
        self.priorities[indices] = priorities
        self.count = min(self.memory_size, self.count + data_size)
        self.current = (self.current + data_size) % self.memory_size

    def sample(self, count):
        """
        Returns count randomly sampled memories
        :param count:
        :return:
        """
        if self.count <= 0:
            raise Exception('Not enough memories in buffer')
        priorities = self.priorities[:self.count, ...]
        priorities = (priorities - np.min(priorities)) / (np.max(priorities) - np.min(priorities))
        normalized_priorities = priorities / np.sum(priorities)
        all_indices = list(range(self.count))

        indices = np.random.choice(all_indices, size=count, p=normalized_priorities)
        return [x[indices] for x in self.buffers]

    def __get_insertion_indices(self, data_size):
        return self.__get_contiguous_indices(self.current, data_size)

    def __get_contiguous_indices(self, start, count):
        if start + count > self.memory_size:
            return list(range(start, self.memory_size)) + list(range(count-(self.memory_size - start)))
        else:
            return list(range(start, start + count))
