import unittest
import numpy as np
from replay_memory import ReplayMemory, PrioritizedReplayMemory, get_buffers_for_arrays
from circular_buffer_index import CircularBufferIndex


class TestReplayMemory(unittest.TestCase):

    def test_replay_memory(self):
        replay_memory = ReplayMemory([np.empty((10, 3, 3, 2)), np.empty(10)])
        replay_memory.add([np.ones((5, 3, 3, 2)), np.ones(5)])
        self.assertEqual(replay_memory.count, 5)
        self.assertEqual(np.sum(replay_memory.buffers[0][0, ...]), 3 * 3 * 2)

        replay_memory.add([np.zeros((1, 3, 3, 2)), np.zeros(1)])
        self.assertEqual(np.sum(replay_memory.buffers[0][5]), 0)
        self.assertEqual(replay_memory.current, 6)

        replay_memory.add([np.zeros((6, 3, 3, 2)), np.zeros(6)])
        self.assertEqual(replay_memory.current, 2)
        self.assertEqual(replay_memory.count, 10)

        sums = [np.sum(x) for x in replay_memory.buffers[0]]
        expected_sums = [0.0, 0.0, 18.0, 18.0, 18.0, 0.0, 0.0, 0.0, 0.0, 0.0]
        self.assertEqual(sums, expected_sums)

        replay_memory.add([np.zeros((10, 3, 3, 2)), np.zeros(10)])
        self.assertEqual(np.sum(replay_memory.buffers[0]), 0)

        replay_memory.add([np.ones((6, 3, 3, 2)), np.ones(6)])
        replay_memory.add([np.zeros((4, 3, 3, 2)), np.ones(4)])

        last_6 = replay_memory.sample_last_contiguous(6, 4)
        self.assertEqual(np.mean(last_6[0]), 1.0)
        last_4 = replay_memory.sample_last_contiguous(4, 0)
        self.assertEqual(np.mean(last_4[0]), 0.0)


        buffer_size = 20
        shapes = [(1,2,3), (3,4,5), (1,1,3)]
        expected_shapes = [(buffer_size, 1,2,3), (buffer_size, 3,4,5), (buffer_size, 1,1,3)]
        arrays = [np.ones(s, dtype=np.float16) for s in shapes]
        expected = [np.empty(s, dtype=np.float16) for s in expected_shapes]
        buffers = get_buffers_for_arrays(arrays, buffer_size)
        for buffer, e_buffer in zip(buffers, expected):
            self.assertEqual(buffer.shape, e_buffer.shape)
            self.assertEqual(buffer.dtype, e_buffer.dtype)

    def test_prioritized_replay_memory(self):
        buffers = [np.arange(32, dtype=np.float32).reshape((32, 1)), np.ones((32, 2, 2), dtype=np.float32)]
        priorities = np.arange(32, dtype=np.float32)
        priorities /= np.sum(priorities)
        replay_memory = PrioritizedReplayMemory([np.empty((50, 1)), np.empty((50, 2, 2))])
        replay_memory.add(buffers, priorities)
        sampled = replay_memory.sample(10000)
        s = sampled[0]
        for i in range(32):
            found = s[s == float(i)].shape[0]
            expected = int(priorities[i] * 10000)
            # print(found, expected)
            self.assertTrue(abs(found - expected) <= float(expected)/2.0)


    def test_circular_buffer_index(self):
        buffer = np.zeros(10)
        to_add = list(range(150))
        to_add_idx = 0
        circular_buffer_index = CircularBufferIndex(10)
        for i in range(20):
            buffer[circular_buffer_index.index()] = to_add[to_add_idx]
            to_add_idx += 1
            circular_buffer_index.increase()

        expected_last_10 = np.array([10, 11, 12, 13, 14, 15, 16, 17, 18, 19])
        np.testing.assert_array_equal(buffer[circular_buffer_index.get_all_indices()], expected_last_10)
        expected_oldest_5 = np.array([10, 11, 12, 13, 14])
        np.testing.assert_array_equal(buffer[circular_buffer_index.get_n_oldest_indices(5)], expected_oldest_5)
        expected_newest_5 = np.array([15, 16, 17, 18, 19])
        np.testing.assert_array_equal(buffer[circular_buffer_index.get_n_newest_indices(5)], expected_newest_5)

if __name__ == '__main__':
    unittest.main()

def random_feed():
    self = None
    num_actions=2
    rnd_states = np.random.randint(0, 255, (1, 84, 84, 4), dtype=np.uint32)
    rnd_actions = np.eye(num_actions)[np.random.randint(0, num_actions, 1)]
    rnd_actor_targets = np.random.random(1).astype(np.float32)
    rnd_qrets = np.random.random(1).astype(np.float32)
    rnf_imp_sampl = np.ones(1, dtype=np.float32)

    feed_dict={self.network.input_ph: rnd_states,
               self.trpo_network.input_ph: rnd_states,
               self.network.actor_critic_term: rnd_actor_targets,
               self.trpo_network.actor_critic_term: rnd_actor_targets,
               self.network.q_rets: rnd_qrets,
               self.trpo_network.q_rets: rnd_qrets,
               self.network.selected_action_ph: rnd_actions,
               self.trpo_network.selected_action_ph: rnd_actions,
               self.network.importance_sampling_ph: rnf_imp_sampl,
               self.trpo_network.importance_sampling_ph: rnf_imp_sampl,
               self.learning_rate: self.get_lr()}



